local _floor, _min, _max, _concat = math.floor, math.min, math.max, table.concat
local _textsize, _fillrect, _drawline, _drawtext = gfx.textSize, graphics.fillRect, graphics.drawLine, graphics.drawText
local _gettime = socket.gettime
local _band, _blshift, _brshift = bit.band, bit.lshift, bit.rshift

local CONNECT_IP = "lbphacker.pw"
local CONNECT_PORT = 56628
local CONNECT_TIMEOUT = 1

local WIDTH = 215
local HEIGHT = gfx.HEIGHT
local HISTORYHEIGHT = HEIGHT - 43
local ROWHEIGHT = select(2, _textsize("\n")) - select(2, _textsize(""))
local GAPFIX = select(2, _textsize("\n")) - 2 * select(2, _textsize(""))
local ROWS = _floor(HISTORYHEIGHT / ROWHEIGHT)
local BORDER = _floor(ROWHEIGHT / 2)
local MSGGAP = _floor(ROWHEIGHT / 3)
local SCROLLBARWIDTH = 6
local SCROLLSPEED = 20
local REPEATTIMEOUT = 400
local REPEATINTERVAL = 150
local MESSAGEVIEWTIMEOUT = 10
local MESSAGEFADETIME = 2
local MAXBUFFERLEN = 1000
local INPUTPOS = _floor((HEIGHT - HISTORYHEIGHT) / 2) + HISTORYHEIGHT - 3
local WRAPWIDTH = WIDTH - 2 * BORDER
local WRAPWIDTHBORDER = WIDTH - 2 * BORDER - SCROLLBARWIDTH

local showingPopup = false
local scroll_offset
local view_inner = {}
local view_outer = {}

local server_connection, still_connecting

local function num2le(size, num)
	local result = ""
	for ix = 0, size - 1 do
		result = result .. string.char(_band(_brshift(num, ix * 8), 0xFF))
	end
	return result
end

local function le2num(size, str)
	local result = 0
	for ix = 0, size - 1 do
		result = result + _blshift(str:byte(ix + 1), ix * 8)
	end
	return result
end

local function set_view_scroll(view, scroll)
	local scroll_bottom = scroll + HISTORYHEIGHT - 2 * BORDER
	view.first[view].partial = false
	view.last[view].partial = false
	while view.first[view].top > scroll and view.first.prev do view.first = view.first.prev end
	while view.first[view].bottom < scroll and view.first.next do view.first = view.first.next end
	while view.last[view].bottom < scroll_bottom and view.last.next do view.last = view.last.next end
	while view.last[view].top > scroll_bottom and view.last.prev do view.last = view.last.prev end
	view.first[view].partial = true
	view.last[view].partial = true
	view.scroll = scroll
end

local function fully_visible(a, b)
	return BORDER <= a and HISTORYHEIGHT - BORDER >= b
end

local function singlewrap_reverse(source, ww)
	local result = ""
	for letter in source:reverse():gmatch(".") do
		local trythis = letter .. result
		if _textsize(trythis) > ww then
			break
		end
		result = trythis
	end
	return result
end

local buffer_str, buffer_str_parsed, buffer_str_ok, buffer_len, buffer_vstr, buffer_vstrpos
local function buffer_set(str)
	buffer_str = str
	buffer_str_ok, buffer_str_parsed = pcall(function()
		return loadstring("return \"" .. buffer_str:gsub("\"", "\\\"") .. "\"")()
	end)
	if not buffer_str_ok then
		buffer_str_parsed = buffer_str
	end
	buffer_vstr = singlewrap_reverse(buffer_str, WRAPWIDTH)
	buffer_vstrpos = WIDTH - BORDER - _textsize(buffer_vstr)
end
local function buffer_reset()
	buffer_len = 0
	buffer_set("")
end

local last_y = 0
local MAXSCROLL
local last_message
local function add_message(msgtext, system)
	local CURRENTCOLOR = "\015\255\255\255"
	local sanelines = {CURRENTCOLOR}
	local linecounter = 1
	local wbuffer = ""
	local parsingcolor = 0
	for letter in (msgtext .. " "):gmatch(".") do
		if parsingcolor > 0 then
			CURRENTCOLOR = CURRENTCOLOR .. letter
			parsingcolor = parsingcolor + 1
			if parsingcolor == 4 then
				parsingcolor = 0
				wbuffer = wbuffer .. CURRENTCOLOR
			end
		elseif letter == "\015" then
			parsingcolor = 1
			CURRENTCOLOR = "\015"
		elseif letter == " " or letter == "\n" then
			local trythis = sanelines[linecounter] .. wbuffer
			if _textsize(trythis) > WRAPWIDTHBORDER then
				linecounter = linecounter + 1
				sanelines[linecounter] = CURRENTCOLOR .. wbuffer .. " "
			else
				sanelines[linecounter] = trythis .. " "
			end
			if letter == "\n" then
				linecounter = linecounter + 1
				sanelines[linecounter] = CURRENTCOLOR
			end
			wbuffer = ""
		else
			wbuffer = wbuffer .. letter
		end
	end
	
	local height = MSGGAP + linecounter * ROWHEIGHT
	local message = {
		lines = sanelines,
		lines_len = #sanelines,
		fulltext = _concat(sanelines, "\n"),
		system = system,
		[view_inner] = {
			top = last_y,
			bottom = last_y + height - 1
		},
		[view_outer] = {
			top = last_y,
			bottom = last_y + height - 1
		},
		fadetime = _gettime() + MESSAGEVIEWTIMEOUT + MESSAGEFADETIME
	}
	last_y = last_y + height
	MAXSCROLL = _max(last_y - HISTORYHEIGHT + 2 * BORDER, 0)
	
	if last_message then
		last_message.next = message
		message.prev = last_message
	else
		view_inner.first = message
		view_inner.last = message
		view_outer.first = message
		view_outer.last = message
		message.fadetime = 0
	end
	last_message = message
	
	set_view_scroll(view_inner, MAXSCROLL)
	set_view_scroll(view_outer, MAXSCROLL)
end

local function sysmessage(msg)
	add_message("\015\255\255\001" .. msg)
end

local commands
commands = {
	help = function()
		sysmessage([[Available commands:
 - /help
 - /connect [server:port]
 - /disconnect]])
	end,
	connect = function(words)
		if server_connection then
			sysmessage("Already connected")
		else
			local ok, ip, port
			ok, server_connection = pcall(function()
				if words[1] then
					ip, port = words[1]:match("(.+):(.+)")
				else
					ip, port = CONNECT_IP, CONNECT_PORT
				end
				if not ip then return end
				local server_connection = socket.tcp()
				server_connection:settimeout(0)
				server_connection:connect(ip, tonumber(port))
				still_connecting = socket.gettime() + CONNECT_TIMEOUT
				return server_connection
			end)
		end
	end,
	nick = function(words)
		if not words[1] then
			sysmessage("Missing arguments")
			return
		end
		if server_connection then
			server_connection:send(num2le(1, 4) .. num2le(2, #words[1]) .. words[1])
		else
			sysmessage("Not connected")
		end
	end,
	disconnect = function()
		if server_connection then
			server_connection:send(num2le(1, 1))
			server_connection = nil
			sysmessage("Disconnected")
		else
			sysmessage("Not connected")
		end
	end,
}

local function run_command(buffer_str_parsed)
	buffer_str_parsed = buffer_str_parsed:gsub("^%s*(.+)%s*$", "%1")
	if buffer_str_parsed:sub(1, 1) == "/" then
		local words = {}
		for word in buffer_str_parsed:sub(2):gmatch("%S+") do
			words[#words + 1] = word
		end
		if not words[1] then
			sysmessage("No command parsed.")
		elseif commands[words[1]] then
			local command = words[1]
			table.remove(words, 1)
			commands[command](words)
		else
			sysmessage("Unknown command '" .. words[1] .. "'.")
		end
	else
		if server_connection then
			server_connection:send(num2le(1, 3) .. num2le(2, #buffer_str_parsed) .. buffer_str_parsed)
		end
	end
end

tpt.boost.register("Chat", {
	mouseclick = function(mousex, mousey, button, event, wheel)
		if showingPopup then
			if wheel ~= 0 then
				set_view_scroll(view_inner, _min(_max(view_inner.scroll + wheel * -SCROLLSPEED, 0), MAXSCROLL))
				return false
			end
		else
			if button == 1 and event == 1 and tpt.mousex <= WIDTH then
				local time_now = _gettime()
				
				local message = view_outer.first
				while true do
					if showingPopup or message.fadetime > time_now then
						
						local fully_exposed = not message[view_outer].partial
						local cur_y = message[view_outer].top - view_outer.scroll + BORDER
						
						if fully_exposed then
							if tpt.mousey >= cur_y - 2 and tpt.mousey < cur_y - 2 + ROWHEIGHT * message.lines_len then
								message.fadetime = 0
							end
						else
							for key, value in next, message.lines do
								local posy = cur_y + (key - 1) * ROWHEIGHT
								if fully_visible(posy, posy + ROWHEIGHT) then
									if tpt.mousey >= posy - 2 and tpt.mousey < posy - 2 + ROWHEIGHT then
										message.fadetime = 0
									end
								end
							end
						end
					end
					
					if message == view_outer.last then
						break
					end
					message = message.next
				end
			end
		end
	end,
	step = function()
		local time_now = _gettime()
		
		if server_connection then
			if still_connecting then
				if server_connection:getpeername() then
					local username = tpt.get_name()
					if username == "" then
						username = ("Guest%04i"):format(math.random(0, 9999))
					end
					server_connection:send(num2le(1, 0) .. num2le(2, #username) .. username)
					sysmessage("Connected to " .. CONNECT_IP)
					still_connecting = nil
				end
				if still_connecting and still_connecting < socket.gettime() then
					sysmessage("Failed to connect to " .. CONNECT_IP)
					still_connecting = nil
					server_connection = nil
				end
			else
				while true do
					local packet = server_connection:receive(1)
					if not packet then
						break
					end
					packet = le2num(1, packet)
					if packet == 1 then
						sysmessage("Server closed")
						server_connection:close()
						server_connection = nil
						break
					elseif packet == 3 then
						local msglen = le2num(2, server_connection:receive(2))
						local msg = server_connection:receive(msglen)
						add_message(msg)
					-- else -- ignore
					end
				end
			end
		end
		
		-- if ren.debugHUD() == 1 and tpt.hud() == 1 then
		if tpt.hud() == 1 then
			local view = showingPopup and view_inner or view_outer
			
			if showingPopup then
				_fillrect(0, 0, WIDTH, HEIGHT, 0, 0, 0, 200)
				_drawline(0, HISTORYHEIGHT, WIDTH, HISTORYHEIGHT)
				_drawline(WIDTH, 0, WIDTH, HEIGHT)
			end
			
			local message = view.first
			while true do
				if showingPopup or message.fadetime > time_now then
					local alpha = showingPopup and 1 or _min((message.fadetime - time_now) / MESSAGEFADETIME, 1)
					
					local fully_exposed = not message[view].partial
					local cur_y = message[view].top - view.scroll + BORDER
					local fulltext = message.fulltext
					if fully_exposed then
						if not showingPopup then
							_fillrect(0, cur_y - 2, WIDTH, ROWHEIGHT * message.lines_len, 0, 0, 0, alpha * 200)
						end
						_drawtext(BORDER, cur_y, message.fulltext, 255, 255, 255, alpha * 255)
					else
						for key, value in next, message.lines do
							local posy = cur_y + (key - 1) * ROWHEIGHT
							if fully_visible(posy, posy + ROWHEIGHT) then
								if not showingPopup then
									_fillrect(0, posy - 2, WIDTH, ROWHEIGHT, 0, 0, 0, alpha * 200)
								end
								_drawtext(BORDER, posy, value, 255, 255, 255, alpha * 255)
							end
						end
					end
				end
				
				if message == view.last then
					break
				end
				message = message.next
			end
			
			if showingPopup then -- * TODO: turn this into an actual UI element
				local sb_size = _floor(HISTORYHEIGHT * HISTORYHEIGHT / (MAXSCROLL + HISTORYHEIGHT))
				local sb_pos = _floor((view.scroll / MAXSCROLL) * (HISTORYHEIGHT - sb_size))
				_fillrect(WIDTH - SCROLLBARWIDTH, 0, SCROLLBARWIDTH, HISTORYHEIGHT, 255, 255, 255, 48)
				_fillrect(WIDTH - SCROLLBARWIDTH, sb_pos, SCROLLBARWIDTH, sb_size)
				
				_drawtext(buffer_vstrpos, INPUTPOS, buffer_vstr, 255, buffer_str_ok and 255 or 0, buffer_str_ok and 255 or 0)
			end
		end
	end,
	keypress = function(cchar, code, modkey, state)
		if showingPopup then
			if modkey == 0 and code == 27 and state == 1 then
				showingPopup = false
				tpt.boost.unlock()
			elseif modkey == 0 and code == 13 and state == 1 then
				if buffer_str_parsed ~= "" then
					run_command(buffer_str_parsed)
				end
				showingPopup = false
				tpt.boost.unlock()
			elseif modkey == 0 and code == 8 and state == 1 then
				if buffer_len > 0 then
					buffer_len = buffer_len - 1
					buffer_set(buffer_str:sub(1, buffer_len))
				end
			elseif cchar:byte() > 31 and cchar:byte() < 127 and state == 1 then
				if buffer_len < MAXBUFFERLEN then
					buffer_len = buffer_len + 1
					buffer_set(buffer_str .. cchar)
				end
			end
			return false
		else
			if modkey == 0 and state == 1 and cchar == "t" then
				tpt.boost.lock()
				showingPopup = true
				buffer_reset()
				set_view_scroll(view_inner, MAXSCROLL)
				return false
			end
		end
	end
})

sysmessage("Boost::Chat by LBPHacker, 2016. Type \"/help\" for help.")

do
	local autorun_func, err = loadfile("Chat.Autorun.lua")
	if autorun_func then
		local ok, err = pcall(autorun_func, {
			print = add_message,
			system_message = sysmessage,
			execute = run_command
		})
		if not ok then
			sysmessage("\15\255\1\1" .. err)
		end
	elseif not err:find("cannot open") then
		sysmessage("\15\255\1\1" .. err)
	end
end

